import re
from functools import partial
from typing import Any, Optional

import httpx
import respx

from openai.pagination import SyncPage
from openai.types.file_object import FileObject
from openai.types.file_deleted import FileDeleted

from ._base import StatefulMock, CallContainer
from ..decorators import side_effect
from ..state import StateStore
from ..utils import model_dict, utcnow_unix_timestamp_s


class FilesMock(StatefulMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + "/files"
        self.create = CallContainer()
        self.list = CallContainer()
        self.retrieve = CallContainer()
        self.delete = CallContainer()

    def _register_routes(self, **common: Any) -> None:
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )

        self.list.route = respx.get(url__regex=self.url).mock(
            side_effect=partial(self._list, **common)
        )

        self.retrieve.route = respx.get(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._retrieve, **common)
        )

        self.delete.route = respx.delete(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._delete, **common)
        )

    def __call__(
        self,
        *,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
        state_store: Optional[StateStore] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                latency=latency or 0,
                failures=failures or 0,
                state_store=kwargs["used_state"],
            )

        return self._make_decorator("files_mock", getter, state_store or StateStore())

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route

        filename = ""
        purpose = "assistants"

        content = request.content.decode("utf-8")

        prog = re.compile(
            r'Content-Disposition: form-data;[^;]+; name="purpose"\r\n\r\n(?P<purpose_value>[^\r\n]+)|filename="(?P<filename>[^"]+)"'
        )
        matches = prog.finditer(content)
        for match in matches:
            if match.group("filename"):
                filename = match.group("filename")
            if match.group("purpose_value"):
                purpose = match.group("purpose_value")

        obj = FileObject(
            id=self._faker.file.id(),
            bytes=0,
            created_at=utcnow_unix_timestamp_s(),
            filename=filename,
            object="file",
            purpose=purpose,  # type: ignore
            status="uploaded",
        )
        state_store.files.put(obj)

        return httpx.Response(status_code=201, json=model_dict(obj))

    @side_effect
    def _list(
        self,
        request: httpx.Request,
        route: respx.Route,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.list.route = route

        purpose = request.url.params.get("purpose")
        files = SyncPage[FileObject](
            object="list",
            data=state_store.files.list(purpose=purpose),
        )

        return httpx.Response(status_code=200, json=model_dict(files))

    @side_effect
    def _retrieve(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.retrieve.route = route

        *_, id = request.url.path.split("/")
        file = state_store.files.get(id)

        if not file:
            return httpx.Response(status_code=404)

        return httpx.Response(status_code=200, json=model_dict(file))

    @side_effect
    def _delete(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.delete.route = route

        *_, id = request.url.path.split("/")
        deleted = state_store.files.delete(id)

        return httpx.Response(
            status_code=200,
            json=model_dict(FileDeleted(id=id, deleted=deleted, object="file")),
        )



import time
import warnings
from typing import Any, Callable

import httpx
import respx
from decorator import decorator


@decorator
def side_effect(fn: Callable[..., httpx.Response], *args: Any, **kwargs: Any):
    # inject latency
    latency: float = kwargs.get("latency", 0.0)
    time.sleep(latency)

    try:
        route = next((arg for arg in args if isinstance(arg, respx.Route)))
        failures: int = kwargs.get("failures", 0)
        if route.call_count < failures:
            return httpx.Response(status_code=500)
    except StopIteration:
        warnings.warn("Could not find route in side effect call")

    return fn(*args, **kwargs)


def unwrap(wrapped: Callable[..., Any]) -> Callable[..., Any]:
    """
    https://stackoverflow.com/a/77694433
    """
    closure = wrapped.__closure__
    if closure:
        for cell in closure:
            if hasattr(cell.cell_contents, "__module__"):
                if cell.cell_contents.__module__.split(".")[0] == "openai_responses":
                    continue
            if hasattr(cell.cell_contents, "__closure__"):
                return (
                    cell.cell_contents
                    if cell.cell_contents.__closure__ is None
                    else unwrap(cell.cell_contents)
                )
    return wrapped



import inspect
from functools import wraps
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Protocol

import httpx
import respx
from faker import Faker
from faker_openai_api_provider import OpenAiApiProvider

from ..decorators import unwrap
from ..state import StateStore


class KwargsGetterProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Dict[str, Any]: ...


class Mock(ABC):
    BASE_URL = "https://api.openai.com/v1"

    def __init__(self) -> None:
        _fake = Faker()
        _fake.add_provider(OpenAiApiProvider)
        self._faker: OpenAiApiProvider.Api = _fake.openai()

    @abstractmethod
    def _register_routes(self, **common: Any) -> None: ...


def sort_routes(routes: List[respx.Route]) -> None:
    routes.sort(key=lambda r: len(repr(r._pattern)), reverse=True)  # type: ignore


class StatelessMock(Mock):
    def _make_decorator(
        self,
        mocker_class: str,
        common_kwargs_getter: KwargsGetterProtocol,
    ):
        def decorator(fn: Callable[..., Any]):
            is_async = inspect.iscoroutinefunction(fn)
            argspec = inspect.getfullargspec(fn)
            needs_ref = mocker_class in argspec.args

            async def async_wrapper(*args: Any, **kwargs: Any):
                if needs_ref:
                    kwargs[mocker_class] = self
                common_kwargs = common_kwargs_getter(*args, **kwargs)
                with respx.mock:
                    self._register_routes(**common_kwargs)
                    sort_routes(respx.mock.routes._routes)
                    return await fn(*args, **kwargs)

            def wrapper(*args: Any, **kwargs: Any):
                if needs_ref:
                    kwargs[mocker_class] = self
                common_kwargs = common_kwargs_getter(*args, **kwargs)
                with respx.mock:
                    self._register_routes(**common_kwargs)
                    sort_routes(respx.mock.routes._routes)
                    return fn(*args, **kwargs)

            return wrapper if not is_async else async_wrapper

        return decorator


class StatefulMock(Mock):

    def _make_decorator(
        self,
        mocker_class: str,
        common_kwargs_getter: KwargsGetterProtocol,
        state_store: StateStore,
    ):
        def decorator(fn: Callable[..., Any]):

            is_async = inspect.iscoroutinefunction(fn)
            argspec = inspect.getfullargspec(unwrap(fn))
            needs_ref = mocker_class in argspec.args

            @wraps(fn)
            async def async_wrapper(*args: Any, **kwargs: Any):
                if needs_ref:
                    kwargs[mocker_class] = self
                state = kwargs.get("state_store", state_store)
                common_kwargs = common_kwargs_getter(used_state=state, *args, **kwargs)
                with respx.mock:
                    self._register_routes(**common_kwargs)
                    sort_routes(respx.mock.routes._routes)
                    return await fn(*args, **kwargs)

            @wraps(fn)
            def wrapper(*args: Any, **kwargs: Any):
                if needs_ref:
                    kwargs[mocker_class] = self
                state = kwargs.get("state_store", state_store)
                common_kwargs = common_kwargs_getter(used_state=state, *args, **kwargs)
                with respx.mock:
                    self._register_routes(**common_kwargs)
                    sort_routes(respx.mock.routes._routes)
                    return fn(*args, **kwargs)

            return wrapper if not is_async else async_wrapper

        return decorator


class CallContainer:
    def __init__(self) -> None:
        self.route = respx.Route()
        self.request: Optional[httpx.Request] = None
        self.response: Optional[httpx.Response] = None



import json
from functools import partial
from typing import Any, List, Optional

import httpx
import respx

from openai.types.embedding import Embedding
from openai.types.embedding_create_params import EmbeddingCreateParams
from openai.types.create_embedding_response import CreateEmbeddingResponse, Usage

from ._base import StatelessMock, CallContainer
from ..decorators import side_effect
from ..utils import model_dict


class EmbeddingsMock(StatelessMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + "/embeddings"
        self.create = CallContainer()

    def _register_routes(self, **common: Any) -> None:
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )

    def __call__(
        self,
        *,
        embedding: Optional[List[float]] = None,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                embedding=embedding or [],
                latency=latency or 0,
                failures=failures or 0,
            )

        return self._make_decorator("embeddings_mock", getter)

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        embedding: List[float],
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route

        content: EmbeddingCreateParams = json.loads(request.content)

        embeddings = CreateEmbeddingResponse(
            data=[Embedding(embedding=embedding, index=0, object="embedding")],
            model=content["model"],
            object="list",
            usage=Usage(prompt_tokens=0, total_tokens=0),
        )

        return httpx.Response(status_code=201, json=model_dict(embeddings))



from typing import Dict, Generic, List, Literal, Optional, TypeVar, Union

from openai.types import FileObject
from openai.types.beta.assistant import Assistant
from openai.types.beta.thread import Thread
from openai.types.beta.threads.message import Message
from openai.types.beta.threads.run import Run
from openai.types.beta.threads.runs.run_step import RunStep


__all__ = ["StateStore"]

M = TypeVar(
    "M",
    bound=Union[
        FileObject,
        Assistant,
        Thread,
        Message,
        Run,
        RunStep,
    ],
)


class StateStore:
    def __init__(self) -> None:
        self.files = FileStore()
        self.beta = Beta()


class Beta:
    def __init__(self) -> None:
        self.assistants = AssistantStore()
        self.threads = ThreadStore()


class BaseStore(Generic[M]):
    def __init__(self) -> None:
        self._data: Dict[str, M] = {}

    def put(self, obj: M) -> None:
        self._data[obj.id] = obj

    def get(self, id: str) -> Optional[M]:
        return self._data.get(id)

    def delete(self, id: str) -> bool:
        if self._data.get(id):
            del self._data[id]
            return True
        else:
            return False


# TODO: add content storage
class FileStore(BaseStore[FileObject]):
    def list(
        self,
        purpose: Optional[
            Literal[
                "fine-tune",
                "fine-tune-results",
                "assistants",
                "assistants_output",
            ]
        ] = None,
    ) -> List[FileObject]:
        files = list(self._data.values())
        if purpose:
            files = [file for file in files if file.purpose == purpose]
        return files


class AssistantStore(BaseStore[Assistant]):
    def list(
        self,
        limit: Optional[int] = None,
        order: Optional[str] = None,
        after: Optional[str] = None,
        before: Optional[str] = None,
    ) -> List[Assistant]:
        objs = list(self._data.values())
        objs = list(reversed(objs)) if (order or "desc") == "desc" else objs

        start_ix = 0
        if after:
            obj = self._data.get(after)
            if obj:
                start_ix = objs.index(obj) + 1

        end_ix = None
        if before:
            obj = self._data.get(before)
            if obj:
                end_ix = objs.index(obj)

        objs = objs[start_ix:end_ix]
        return objs[: (limit or 20) + 1]


class ThreadStore(BaseStore[Thread]):
    def __init__(self) -> None:
        super().__init__()
        self.messages = MessageStore()
        self.runs = RunStore()


class MessageStore(BaseStore[Message]):
    def list(
        self,
        thread_id: str,
        limit: Optional[int] = None,
        order: Optional[str] = None,
        after: Optional[str] = None,
        before: Optional[str] = None,
    ) -> List[Message]:
        objs = [m for m in list(self._data.values()) if m.thread_id == thread_id]
        objs = list(reversed(objs)) if (order or "desc") == "desc" else objs

        start_ix = 0
        if after:
            obj = self._data.get(after)
            if obj:
                start_ix = objs.index(obj) + 1

        end_ix = None
        if before:
            obj = self._data.get(before)
            if obj:
                end_ix = objs.index(obj)

        objs = objs[start_ix:end_ix]
        return objs[: (limit or 20) + 1]


class RunStore(BaseStore[Run]):
    def __init__(self) -> None:
        super().__init__()
        self.steps = RunStepStore()

    def list(
        self,
        thread_id: str,
        limit: Optional[int] = None,
        order: Optional[str] = None,
        after: Optional[str] = None,
        before: Optional[str] = None,
    ) -> List[Run]:
        objs = [m for m in list(self._data.values()) if m.thread_id == thread_id]
        objs = list(reversed(objs)) if (order or "desc") == "desc" else objs

        start_ix = 0
        if after:
            obj = self._data.get(after)
            if obj:
                start_ix = objs.index(obj) + 1

        end_ix = None
        if before:
            obj = self._data.get(before)
            if obj:
                end_ix = objs.index(obj)

        objs = objs[start_ix:end_ix]
        return objs[: (limit or 20) + 1]


class RunStepStore(BaseStore[RunStep]):
    def list(
        self,
        thread_id: str,
        run_id: str,
        limit: Optional[int] = None,
        order: Optional[str] = None,
        after: Optional[str] = None,
        before: Optional[str] = None,
    ) -> List[RunStep]:
        objs = [
            m
            for m in list(self._data.values())
            if m.thread_id == thread_id and m.run_id == run_id
        ]
        objs = list(reversed(objs)) if (order or "desc") == "desc" else objs

        start_ix = 0
        if after:
            obj = self._data.get(after)
            if obj:
                start_ix = objs.index(obj) + 1

        end_ix = None
        if before:
            obj = self._data.get(before)
            if obj:
                end_ix = objs.index(obj)

        objs = objs[start_ix:end_ix]
        return objs[: (limit or 20) + 1]



import json
from functools import partial
from typing import Any, List, Literal, Optional, TypedDict

import httpx
import respx

from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.completion_usage import CompletionUsage
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage,
    FunctionCall,
)
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.chat.completion_create_params import CompletionCreateParams

from ._base import StatelessMock, CallContainer
from ..decorators import side_effect
from ..utils import model_dict, utcnow_unix_timestamp_s


class ChatMock:
    def __init__(self) -> None:
        self.completions = ChatCompletionMock()


class PartialFunctionCall(TypedDict, total=False):
    arguments: str
    name: str


class PartialToolCall(TypedDict, total=False):
    function: PartialFunctionCall


class PartialMessage(TypedDict, total=False):
    content: str
    function_call: PartialFunctionCall
    tool_calls: List[PartialToolCall]


class PartialChoice(TypedDict, total=False):
    finish_reason: Literal[
        "stop",
        "length",
        "tool_calls",
        "content_filter",
        "function_call",
    ]
    message: PartialMessage


class ChatCompletionMock(StatelessMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + "/chat/completions"
        self.create = CallContainer()

    def _register_routes(self, **common: Any) -> None:
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )

    def __call__(
        self,
        *,
        choices: Optional[List[PartialChoice]] = None,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                choices=choices or [],
                latency=latency or 0,
                failures=failures or 0,
            )

        return self._make_decorator("chat_completion_mock", getter)

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        choices: List[PartialChoice],
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route

        content: CompletionCreateParams = json.loads(request.content)

        completion = ChatCompletion(
            id=self._faker.chat.completion.id(),
            choices=[
                self._choice_partial_to_model(i, p) for i, p in enumerate(choices)
            ],
            model=content["model"],
            created=utcnow_unix_timestamp_s(),
            system_fingerprint="",
            object="chat.completion",
            usage=CompletionUsage(completion_tokens=0, prompt_tokens=0, total_tokens=0),
        )

        return httpx.Response(status_code=201, json=model_dict(completion))

    def _choice_partial_to_model(self, i: int, p: PartialChoice) -> Choice:

        def fn_call_partial_to_model(p: Optional[PartialFunctionCall]):
            if p is None:
                return None
            else:
                return FunctionCall(
                    arguments=p.get("arguments", ""),
                    name=p.get("name", ""),
                )

        def tool_calls_partial_to_model(p: Optional[List[PartialToolCall]]):
            if p is None:
                return None
            else:
                calls: List[ChatCompletionMessageToolCall] = []
                for partial_call in p:
                    call_function = partial_call.get("function", {})
                    calls.append(
                        ChatCompletionMessageToolCall(
                            id=self._faker.beta.thread.run.step.step_details.tool_call.id(),
                            function=Function(
                                arguments=call_function.get("arguments", ""),
                                name=call_function.get("name", ""),
                            ),
                            type="function",
                        )
                    )

        message = ChatCompletionMessage(
            content=p.get("message", {}).get("content"),
            role="assistant",
            function_call=fn_call_partial_to_model(
                p.get("message", {}).get("function_call")
            ),
            tool_calls=tool_calls_partial_to_model(
                p.get("message", {}).get("tool_calls")
            ),
        )

        return Choice(
            finish_reason=p.get("finish_reason", "stop"),
            message=message,
            index=i,
        )



import json
from functools import partial
from typing import Any, Iterable, List, Optional

import httpx
import respx

from openai.types import FunctionDefinition
from openai.pagination import SyncCursorPage
from openai.types.beta.assistant import Assistant
from openai.types.beta.function_tool import FunctionTool
from openai.types.beta.retrieval_tool import RetrievalTool
from openai.types.beta.assistant_tool import AssistantTool
from openai.types.beta.assistant_deleted import AssistantDeleted
from openai.types.beta.code_interpreter_tool import CodeInterpreterTool
from openai.types.beta.assistant_tool_param import AssistantToolParam
from openai.types.beta.assistant_create_params import AssistantCreateParams
from openai.types.beta.assistant_update_params import AssistantUpdateParams

from ._base import StatefulMock, CallContainer
from ..decorators import side_effect
from ..state import StateStore
from ..utils import model_dict, utcnow_unix_timestamp_s


class AssistantsMock(StatefulMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + "/assistants"
        self.create = CallContainer()
        self.list = CallContainer()
        self.retrieve = CallContainer()
        self.update = CallContainer()
        self.delete = CallContainer()

    def _register_routes(self, **common: Any) -> None:
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )
        self.list.route = respx.get(url__regex=self.url).mock(
            side_effect=partial(self._list, **common)
        )
        self.retrieve.route = respx.get(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._retrieve, **common)
        )
        self.update.route = respx.post(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._update, **common)
        )
        self.delete.route = respx.delete(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._delete, **common)
        )

    def __call__(
        self,
        *,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
        state_store: Optional[StateStore] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                latency=latency or 0,
                failures=failures or 0,
                state_store=kwargs["used_state"],
            )

        return self._make_decorator(
            "assistants_mock", getter, state_store or StateStore()
        )

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route

        content: AssistantCreateParams = json.loads(request.content)

        content_tools = content.get("tools", [])

        asst = Assistant(
            id=self._faker.beta.assistant.id(),
            created_at=utcnow_unix_timestamp_s(),
            description=content.get("description"),
            file_ids=content.get("file_ids", []),
            instructions=content.get("instructions"),
            metadata=content.get("metadata"),
            model=content["model"],
            name=content.get("name"),
            object="assistant",
            tools=self._parse_tool_params(content_tools),
        )
        state_store.beta.assistants.put(asst)

        return httpx.Response(status_code=201, json=model_dict(asst))

    @side_effect
    def _list(
        self,
        request: httpx.Request,
        route: respx.Route,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.list.route = route

        limit = request.url.params.get("limit")
        order = request.url.params.get("order")
        after = request.url.params.get("after")
        before = request.url.params.get("before")

        assts = SyncCursorPage[Assistant](
            data=state_store.beta.assistants.list(limit, order, after, before)
        )

        return httpx.Response(status_code=200, json=model_dict(assts))

    @side_effect
    def _retrieve(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.retrieve.route = route

        *_, id = request.url.path.split("/")
        asst = state_store.beta.assistants.get(id)

        if not asst:
            return httpx.Response(status_code=404)

        else:
            return httpx.Response(status_code=200, json=model_dict(asst))

    @side_effect
    def _update(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.update.route = route

        *_, id = request.url.path.split("/")
        content: AssistantUpdateParams = json.loads(request.content)

        asst = state_store.beta.assistants.get(id)

        if not asst:
            return httpx.Response(status_code=404)

        asst.description = content.get("description", asst.description)
        asst.file_ids = content.get("file_ids", asst.file_ids)
        asst.instructions = content.get("instructions", asst.instructions)
        asst.metadata = content.get("metadata", asst.metadata)
        asst.model = content.get("model", asst.model)
        asst.name = content.get("name", asst.name)

        update_tools = content.get("tools")
        if update_tools:
            asst.tools = self._parse_tool_params(update_tools)

        state_store.beta.assistants.put(asst)

        return httpx.Response(status_code=200, json=model_dict(asst))

    @side_effect
    def _delete(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.delete.route = route

        *_, id = request.url.path.split("/")
        deleted = state_store.beta.assistants.delete(id)

        return httpx.Response(
            status_code=200,
            json=model_dict(
                AssistantDeleted(id=id, deleted=deleted, object="assistant.deleted")
            ),
        )

    @staticmethod
    def _parse_tool_params(
        tool_params: Iterable[AssistantToolParam],
    ) -> List[AssistantTool]:
        tools: List[AssistantTool] = []

        for tool in tool_params:
            if tool["type"] == "code_interpreter":
                tools.append(CodeInterpreterTool(type=tool["type"]))
            elif tool["type"] == "retrieval":
                tools.append(RetrievalTool(type=tool["type"]))
            else:
                tools.append(
                    FunctionTool(
                        type=tool["type"],
                        function=FunctionDefinition(
                            name=tool["function"]["name"],
                            description=tool["function"].get("description"),
                            parameters=tool["function"].get("parameters"),
                        ),
                    )
                )

        return tools



from openai_responses.endpoints.assistants import AssistantsMock
from openai_responses.endpoints.chat import ChatMock, ChatCompletionMock
from openai_responses.endpoints.embeddings import EmbeddingsMock
from openai_responses.endpoints.files import FilesMock
from openai_responses.endpoints.threads import (
    ThreadsMock,
    MessagesMock,
    RunsMock,
    RunStepsMock,
)

__all__ = [
    # main API
    "mock",
    # mockers
    "AssistantsMock",
    "ChatCompletionMock",
    "EmbeddingsMock",
    "FilesMock",
    "ThreadsMock",
    "MessagesMock",
    "RunsMock",
    "RunStepsMock",
]


class Api:
    def __init__(self) -> None:
        self.files = FilesMock()
        self.embeddings = EmbeddingsMock()
        self.chat = ChatMock()
        self.beta = Beta()


class Beta:
    def __init__(self) -> None:
        self.assistants = AssistantsMock()
        self.threads = ThreadsMock()


mock = Api()



import json
from functools import partial
from typing import Any, List, Literal, Optional, Sequence, TypedDict, Union

import httpx
import respx

from openai.pagination import SyncCursorPage
from openai.types.beta.assistant import Assistant

from openai.types.beta.thread import Thread
from openai.types.beta.thread_deleted import ThreadDeleted
from openai.types.beta.thread_update_params import ThreadUpdateParams
from openai.types.beta.thread_create_params import (
    ThreadCreateParams,
    Message as ThreadMessageCreateParams,
)

from openai.types.beta.threads.text import Text
from openai.types.beta.threads.text_content_block import TextContentBlock

from openai.types.beta.threads.message import Message
from openai.types.beta.threads.message_create_params import MessageCreateParams
from openai.types.beta.threads.message_update_params import MessageUpdateParams

from openai.types.beta.threads.run import Run, LastError, RequiredAction, Usage
from openai.types.beta.threads.run_create_params import RunCreateParams
from openai.types.beta.threads.run_update_params import RunUpdateParams

from openai.types.beta.threads.runs.run_step import (
    RunStep,
    LastError as RunStepLastError,
)
from openai.types.beta.threads.runs.tool_calls_step_details import ToolCallsStepDetails
from openai.types.beta.threads.runs.message_creation_step_details import (
    MessageCreationStepDetails,
)

from ._base import StatefulMock, CallContainer
from ._partial_schemas import PartialRun, PartialRunStep
from .assistants import AssistantsMock
from ..decorators import side_effect
from ..state import StateStore
from ..utils import model_dict, model_parse, utcnow_unix_timestamp_s

__all__ = ["ThreadsMock", "MessagesMock", "RunsMock"]


class ThreadsMock(StatefulMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + "/threads"
        self.create = CallContainer()
        self.retrieve = CallContainer()
        self.update = CallContainer()
        self.delete = CallContainer()

        self.messages = MessagesMock()
        self.runs = RunsMock()

    def _register_routes(self, **common: Any) -> None:
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )
        self.retrieve.route = respx.get(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._retrieve, **common)
        )
        self.update.route = respx.post(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._update, **common)
        )
        self.delete.route = respx.delete(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._delete, **common)
        )

    def __call__(
        self,
        *,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
        state_store: Optional[StateStore] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                latency=latency or 0,
                failures=failures or 0,
                state_store=kwargs["used_state"],
            )

        return self._make_decorator("threads_mock", getter, state_store or StateStore())

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route

        content: ThreadCreateParams = json.loads(request.content)

        thread = Thread(
            id=self._faker.beta.thread.id(),
            created_at=utcnow_unix_timestamp_s(),
            metadata=content.get("metadata"),
            object="thread",
        )
        messages = [
            self.messages._parse_message_create_params(thread.id, m)
            for m in content.get("messages", [])
        ]

        state_store.beta.threads.put(thread)
        for message in messages:
            state_store.beta.threads.messages.put(message)

        return httpx.Response(status_code=201, json=model_dict(thread))

    @side_effect
    def _retrieve(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.retrieve.route = route

        *_, id = request.url.path.split("/")
        thread = state_store.beta.threads.get(id)

        if not thread:
            return httpx.Response(status_code=404)

        else:
            return httpx.Response(status_code=200, json=model_dict(thread))

    @side_effect
    def _update(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.update.route = route

        *_, id = request.url.path.split("/")
        content: ThreadUpdateParams = json.loads(request.content)

        thread = state_store.beta.threads.get(id)

        if not thread:
            return httpx.Response(status_code=404)

        thread.metadata = content.get("metadata", thread.metadata)

        state_store.beta.threads.put(thread)

        return httpx.Response(status_code=200, json=model_dict(thread))

    @side_effect
    def _delete(
        self,
        request: httpx.Request,
        route: respx.Route,
        id: str,
        state_store: StateStore,
        **kwargs: Any,
    ) -> httpx.Response:
        self.delete.route = route

        *_, id = request.url.path.split("/")
        deleted = state_store.beta.threads.delete(id)

        return httpx.Response(
            status_code=200,
            json=model_dict(
                ThreadDeleted(id=id, deleted=deleted, object="thread.deleted")
            ),
        )


class MessagesMock(StatefulMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + r"/threads/(?P<thread_id>\w+)/messages"
        self.create = CallContainer()
        self.list = CallContainer()
        self.retrieve = CallContainer()
        self.update = CallContainer()

    def _register_routes(self, **common: Any) -> None:
        self.retrieve.route = respx.get(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._retrieve, **common)
        )
        self.update.route = respx.post(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._update, **common)
        )
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )
        self.list.route = respx.get(url__regex=self.url).mock(
            side_effect=partial(self._list, **common)
        )

    def __call__(
        self,
        *,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
        state_store: Optional[StateStore] = None,
        validate_thread_exists: Optional[bool] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                latency=latency or 0,
                failures=failures or 0,
                state_store=kwargs["used_state"],
                validate_thread_exists=validate_thread_exists or False,
            )

        return self._make_decorator(
            "messages_mock", getter, state_store or StateStore()
        )

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        state_store: StateStore,
        validate_thread_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        content: MessageCreateParams = json.loads(request.content)
        message = self._parse_message_create_params(thread_id, content)

        state_store.beta.threads.messages.put(message)

        return httpx.Response(status_code=201, json=model_dict(message))

    @side_effect
    def _list(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        state_store: StateStore,
        validate_thread_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.list.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        limit = request.url.params.get("limit")
        order = request.url.params.get("order")
        after = request.url.params.get("after")
        before = request.url.params.get("before")

        messages = SyncCursorPage[Message](
            data=state_store.beta.threads.messages.list(
                thread_id,
                limit,
                order,
                after,
                before,
            )
        )

        return httpx.Response(status_code=200, json=model_dict(messages))

    @side_effect
    def _retrieve(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        id: str,
        state_store: StateStore,
        validate_thread_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.retrieve.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        *_, id = request.url.path.split("/")
        message = state_store.beta.threads.messages.get(id)

        if not message:
            return httpx.Response(status_code=404)

        else:
            return httpx.Response(status_code=200, json=model_dict(message))

    @side_effect
    def _update(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        id: str,
        state_store: StateStore,
        validate_thread_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.update.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        *_, id = request.url.path.split("/")
        content: MessageUpdateParams = json.loads(request.content)

        message = state_store.beta.threads.messages.get(id)

        if not message:
            return httpx.Response(status_code=404)

        message.metadata = content.get("metadata", message.metadata)

        state_store.beta.threads.messages.put(message)

        return httpx.Response(status_code=200, json=model_dict(message))

    def _parse_message_create_params(
        self,
        thread_id: str,
        create_message: Union[ThreadMessageCreateParams, MessageCreateParams],
    ) -> Message:
        return Message(
            id=self._faker.beta.thread.message.id(),
            content=[
                TextContentBlock(
                    text=Text(annotations=[], value=create_message["content"]),
                    type="text",
                )
            ],
            created_at=utcnow_unix_timestamp_s(),
            file_ids=create_message.get("file_ids", []),
            object="thread.message",
            role=create_message["role"],
            status="completed",
            thread_id=thread_id,
        )


class MultiMethodSequence(TypedDict, total=False):
    create: Sequence[PartialRun]
    retrieve: Sequence[PartialRun]


class RunsMock(StatefulMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = self.BASE_URL + r"/threads/(?P<thread_id>\w+)/runs"
        self.create = CallContainer()
        self.list = CallContainer()
        self.retrieve = CallContainer()
        self.update = CallContainer()
        self.cancel = CallContainer()
        self.submit_tool_outputs = CallContainer()

        self.steps = RunStepsMock()

    def _register_routes(self, **common: Any) -> None:
        self.submit_tool_outputs.route = respx.post(
            url__regex=self.url + r"/(?P<id>\w+)/submit_tool_outputs"
        ).mock(side_effect=partial(self._submit_tool_outputs, **common))
        self.cancel.route = respx.post(
            url__regex=self.url + r"/(?P<id>\w+)/cancel"
        ).mock(side_effect=partial(self._cancel, **common))
        self.update.route = respx.post(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._update, **common)
        )
        self.retrieve.route = respx.get(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._retrieve, **common)
        )
        self.create.route = respx.post(url__regex=self.url).mock(
            side_effect=partial(self._create, **common)
        )
        self.list.route = respx.get(url__regex=self.url).mock(
            side_effect=partial(self._list, **common)
        )

    def __call__(
        self,
        *,
        sequence: Optional[MultiMethodSequence] = None,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
        state_store: Optional[StateStore] = None,
        validate_thread_exists: Optional[bool] = None,
        validate_assistant_exists: Optional[bool] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                sequence=sequence or {},
                latency=latency or 0,
                failures=failures or 0,
                state_store=kwargs["used_state"],
                validate_thread_exists=validate_thread_exists or False,
                validate_assistant_exists=validate_assistant_exists or False,
            )

        return self._make_decorator("runs_mock", getter, state_store or StateStore())

    @side_effect
    def _create(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        sequence: MultiMethodSequence,
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_assistant_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.create.route = route
        failures: int = kwargs.get("failures", 0)

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        content: RunCreateParams = json.loads(request.content)

        partial_run = (
            self._next_partial_run(sequence, route.call_count, failures, "create") or {}
        )
        if validate_assistant_exists:
            asst = state_store.beta.assistants.get(content["assistant_id"])

            if not asst:
                return httpx.Response(status_code=404)

            partial_run = self._merge_partial_run_with_assistant(partial_run, asst)

        run = Run(
            id=self._faker.beta.thread.run.id(),
            object="thread.run",
            created_at=utcnow_unix_timestamp_s(),
            thread_id=thread_id,
            assistant_id=content["assistant_id"],
            status=partial_run.get("status", "queued"),
            required_action=model_parse(
                RequiredAction,
                partial_run.get("required_action"),
            ),
            last_error=model_parse(
                LastError,
                partial_run.get("last_error"),
            ),
            expires_at=partial_run.get("expires_at"),
            started_at=partial_run.get("started_at"),
            cancelled_at=partial_run.get("cancelled_at"),
            failed_at=partial_run.get("failed_at"),
            completed_at=partial_run.get("completed_at"),
            model=partial_run.get("model", "gpt-3.5-turbo"),
            instructions=partial_run.get("instructions", ""),
            tools=AssistantsMock._parse_tool_params(partial_run.get("tools", [])),
            file_ids=partial_run.get("file_ids", []),
            metadata=content.get("metadata"),
            usage=Usage(
                completion_tokens=0,
                prompt_tokens=0,
                total_tokens=0,
            ),
        )

        state_store.beta.threads.runs.put(run)

        return httpx.Response(status_code=201, json=model_dict(run))

    def _retrieve(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        id: str,
        sequence: MultiMethodSequence,
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_assistant_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.retrieve.route = route
        failures: int = kwargs.get("failures", 0)

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        *_, id = request.url.path.split("/")
        run = state_store.beta.threads.runs.get(id)

        if not run:
            return httpx.Response(status_code=404)

        partial_run = (
            self._next_partial_run(sequence, route.call_count, failures, "retrieve")
            or {}
        )

        if validate_assistant_exists:
            asst = state_store.beta.assistants.get(run.assistant_id)
            if asst:
                partial_run = self._merge_partial_run_with_assistant(partial_run, asst)

        run.status = partial_run.get("status", run.status)
        run.expires_at = partial_run.get("expires_at", run.expires_at)
        run.started_at = partial_run.get("started_at", run.started_at)
        run.cancelled_at = partial_run.get("cancelled_at", run.cancelled_at)
        run.failed_at = partial_run.get("failed_at", run.failed_at)
        run.completed_at = partial_run.get("completed_at", run.completed_at)
        run.model = partial_run.get("model", run.model)
        run.instructions = partial_run.get("instructions", run.instructions)
        run.file_ids = partial_run.get("file_ids", run.file_ids)

        if partial_run.get("required_action"):
            run.required_action = model_parse(
                RequiredAction, partial_run.get("required_action")
            )

        if partial_run.get("last_error"):
            run.last_error = model_parse(LastError, partial_run.get("last_error"))

        if partial_run.get("tools"):
            run.tools = AssistantsMock._parse_tool_params(partial_run.get("tools", []))

        state_store.beta.threads.runs.put(run)

        return httpx.Response(status_code=200, json=model_dict(run))

    def _list(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        sequence: MultiMethodSequence,
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_assistant_exists: bool,
        **kwargs: Any,
    ):
        self.list.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        if validate_assistant_exists:
            # TODO: what should be done here?
            pass

        if sequence:
            # TODO: should there be a method sequence for list?
            pass

        limit = request.url.params.get("limit")
        order = request.url.params.get("order")
        after = request.url.params.get("after")
        before = request.url.params.get("before")

        runs = SyncCursorPage[Run](
            data=state_store.beta.threads.runs.list(
                thread_id, limit, order, after, before
            )
        )

        return httpx.Response(status_code=200, json=model_dict(runs))

    def _update(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        id: str,
        sequence: MultiMethodSequence,
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_assistant_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.update.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        if validate_assistant_exists:
            # TODO: what should be done here?
            pass

        if sequence:
            # TODO: should there be a method sequence for update?
            pass

        *_, id = request.url.path.split("/")
        run = state_store.beta.threads.runs.get(id)

        if not run:
            return httpx.Response(status_code=404)

        content: RunUpdateParams = json.loads(request.content)

        run.metadata = content.get("metadata", run.metadata)

        state_store.beta.threads.runs.put(run)

        return httpx.Response(status_code=200, json=model_dict(run))

    def _cancel(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        id: str,
        sequence: MultiMethodSequence,
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_assistant_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.cancel.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        if validate_assistant_exists:
            # TODO: what should be done here?
            pass

        if sequence:
            # TODO: should there be a method sequence for cancel?
            pass

        *_, id, _ = request.url.path.split("/")
        run = state_store.beta.threads.runs.get(id)

        if not run:
            return httpx.Response(status_code=404)

        run.status = "cancelling"

        state_store.beta.threads.runs.put(run)

        return httpx.Response(status_code=200, json=model_dict(run))

    def _submit_tool_outputs(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        id: str,
        sequence: MultiMethodSequence,
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_assistant_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.submit_tool_outputs.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        if validate_assistant_exists:
            # TODO: what should be done here?
            pass

        if sequence:
            # TODO: should there be a method sequence for submit tools?
            pass

        *_, id, _ = request.url.path.split("/")
        run = state_store.beta.threads.runs.get(id)

        if not run:
            return httpx.Response(status_code=404)

        return httpx.Response(status_code=200, json=model_dict(run))

    @staticmethod
    def _next_partial_run(
        sequence: MultiMethodSequence,
        call_count: int,
        failures: int,
        method: Literal["create", "retrieve"],
    ) -> Optional[PartialRun]:
        used_sequence = sequence.get(method, [])
        net_ix = call_count - failures
        try:
            return used_sequence[net_ix]
        except IndexError:
            return None

    @staticmethod
    def _merge_partial_run_with_assistant(
        run: Optional[PartialRun],
        asst: Assistant,
    ) -> PartialRun:
        if not run:
            return {
                "file_ids": asst.file_ids,
                "instructions": asst.instructions or "",
                "model": asst.model,
                "tools": [model_dict(tool) for tool in asst.tools],  # type: ignore
            }
        else:
            return run | {
                "file_ids": run.get("file_ids", asst.file_ids),
                "instructions": run.get("instructions", asst.instructions or ""),
                "model": run.get("model", asst.model),
                "tools": run.get("tools", [model_dict(tool) for tool in asst.tools]),  # type: ignore
            }


class RunStepsMock(StatefulMock):
    def __init__(self) -> None:
        super().__init__()
        self.url = (
            self.BASE_URL + r"/threads/(?P<thread_id>\w+)/runs/(?P<run_id>\w+)/steps"
        )
        self.list = CallContainer()
        self.retrieve = CallContainer()

    def _register_routes(self, **common: Any) -> None:
        self.retrieve.route = respx.get(url__regex=self.url + r"/(?P<id>\w+)").mock(
            side_effect=partial(self._retrieve, **common)
        )
        self.list.route = respx.get(url__regex=self.url).mock(
            side_effect=partial(self._list, **common)
        )

    def __call__(
        self,
        *,
        steps: Optional[List[PartialRunStep]] = None,
        latency: Optional[float] = None,
        failures: Optional[int] = None,
        state_store: Optional[StateStore] = None,
        validate_thread_exists: Optional[bool] = None,
        validate_run_exists: Optional[bool] = None,
    ):
        def getter(*args: Any, **kwargs: Any):
            return dict(
                steps=steps or [],
                latency=latency or 0,
                failures=failures or 0,
                state_store=kwargs["used_state"],
                validate_thread_exists=validate_thread_exists or False,
                validate_run_exists=validate_run_exists or False,
            )

        return self._make_decorator(
            "run_steps_mock", getter, state_store or StateStore()
        )

    def _list(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        run_id: str,
        steps: List[PartialRunStep],
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_run_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.list.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        assistant_id = ""
        if validate_run_exists:
            run = state_store.beta.threads.runs.get(run_id)

            if not run:
                return httpx.Response(status_code=404)

            assistant_id = run.assistant_id

        self._put_steps_in_store(thread_id, run_id, assistant_id, steps, state_store)

        limit = request.url.params.get("limit")
        order = request.url.params.get("order")
        after = request.url.params.get("after")
        before = request.url.params.get("before")

        return httpx.Response(
            status_code=200,
            json=model_dict(
                SyncCursorPage[RunStep](
                    data=state_store.beta.threads.runs.steps.list(
                        thread_id,
                        run_id,
                        limit,
                        order,
                        after,
                        before,
                    )
                )
            ),
        )

    def _retrieve(
        self,
        request: httpx.Request,
        route: respx.Route,
        thread_id: str,
        run_id: str,
        id: str,
        steps: List[PartialRunStep],
        state_store: StateStore,
        validate_thread_exists: bool,
        validate_run_exists: bool,
        **kwargs: Any,
    ) -> httpx.Response:
        self.list.route = route

        if validate_thread_exists:
            thread = state_store.beta.threads.get(thread_id)

            if not thread:
                return httpx.Response(status_code=404)

        assistant_id = ""
        if validate_run_exists:
            run = state_store.beta.threads.runs.get(run_id)

            if not run:
                return httpx.Response(status_code=404)

            assistant_id = run.assistant_id

        self._put_steps_in_store(thread_id, run_id, assistant_id, steps, state_store)

        *_, id = request.url.path.split("/")
        step = state_store.beta.threads.runs.steps.get(id)

        if not step:
            return httpx.Response(status_code=404)

        return httpx.Response(status_code=200, json=model_dict(step))

    def _put_steps_in_store(
        self,
        thread_id: str,
        run_id: str,
        assistant_id: str,
        steps: List[PartialRunStep],
        state_store: StateStore,
    ) -> None:
        for step in steps:
            step_details = step.get("step_details")
            if not step_details:
                continue

            step_details_model: Any = None
            if step_details["type"] == "message_creation":
                step_details_model = model_parse(
                    MessageCreationStepDetails, step_details
                )
            elif step_details["type"] == "tool_calls":
                step_details_model = model_parse(ToolCallsStepDetails, step_details)

            if not step_details_model:
                continue

            state_store.beta.threads.runs.steps.put(
                RunStep(
                    id=step.get("id", self._faker.beta.thread.run.step.id()),
                    assistant_id=step.get("assistant_id", assistant_id),
                    cancelled_at=step.get("cancelled_at"),
                    completed_at=step.get("completed_at"),
                    created_at=utcnow_unix_timestamp_s(),
                    expired_at=step.get("expired_at"),
                    failed_at=step.get("failed_at"),
                    last_error=model_parse(
                        RunStepLastError,
                        step.get("last_error"),
                    ),
                    object="thread.run.step",
                    run_id=run_id,
                    status=step.get("status", "in_progress"),
                    thread_id=thread_id,
                    type=step_details["type"],
                    step_details=step_details_model,
                )
            )



from typing import Annotated, Literal, List, Optional, TypedDict, Union
from typing_extensions import Required

from openai._utils._transform import PropertyInfo
from openai.types.beta.assistant_tool_param import AssistantToolParam
from openai.types.beta.threads.run_status import RunStatus

__all__ = ["PartialRun", "PartialRunStep"]


class PartialFunctionNoOutput(TypedDict):
    name: str
    arguments: str


class PartialRequiredActionFunctionToolCall(TypedDict):
    id: str
    function: PartialFunctionNoOutput
    type: Literal["function"]


class PartialRequiredActionSubmitToolOutputs(TypedDict):
    tool_calls: List[PartialRequiredActionFunctionToolCall]


class PartialRequiredAction(TypedDict):
    submit_tool_outputs: PartialRequiredActionSubmitToolOutputs
    type: Literal["submit_tool_outputs"]


class PartialLastError(TypedDict, total=False):
    code: Literal["server_error", "rate_limit_exceeded", "invalid_prompt"]
    message: str


class PartialRun(TypedDict, total=False):
    status: RunStatus
    required_action: PartialRequiredAction
    last_error: PartialLastError
    expires_at: int
    started_at: int
    cancelled_at: int
    failed_at: int
    completed_at: int
    model: str
    instructions: str
    tools: List[AssistantToolParam]
    file_ids: List[str]


class PartialCodeInterpreterOutputLogs(TypedDict):
    logs: str
    type: Literal["logs"]


class PartialCodeInterpreterOutputImageImage(TypedDict):
    file_id: str


class PartialCodeInterpreterOutputImage(TypedDict):
    type: Literal["image"]
    image: PartialCodeInterpreterOutputImageImage


class PartialCodeInterpreter(TypedDict):
    input: str
    outputs: List[
        Annotated[
            Union[
                PartialCodeInterpreterOutputLogs,
                PartialCodeInterpreterOutputImage,
            ],
            PropertyInfo(discriminator="type"),
        ]
    ]


class PartialCodeInterpreterToolCall(TypedDict):
    type: Literal["code_interpreter"]
    id: str
    code_interpreter: PartialCodeInterpreter


class PartialRetrievalToolCall(TypedDict):
    type: Literal["retrieval"]
    id: str


class PartialFunction(TypedDict, total=False):
    name: Required[str]
    arguments: Required[str]
    output: Optional[str]


class PartialFunctionToolCall(TypedDict):
    type: Literal["function"]
    id: str
    function: PartialFunction


class PartialToolCallsStepDetails(TypedDict):
    type: Literal["tool_calls"]
    tool_calls: List[
        Annotated[
            Union[
                PartialCodeInterpreterToolCall,
                PartialRetrievalToolCall,
                PartialFunctionToolCall,
            ],
            PropertyInfo(discriminator="type"),
        ]
    ]


class PartialMessageCreation(TypedDict):
    message_id: str


class PartialMessageCreationStepDetails(TypedDict):
    type: Literal["message_creation"]
    message_creation: PartialMessageCreation


class PartialRunStep(TypedDict, total=False):
    id: str
    assistant_id: str
    cancelled_at: Optional[int]
    completed_at: Optional[int]
    expired_at: Optional[int]
    failed_at: Optional[int]
    last_error: Optional[PartialLastError]
    status: Literal["in_progress", "cancelled", "failed", "completed", "expired"]
    step_details: Required[
        Annotated[
            Union[PartialMessageCreationStepDetails, PartialToolCallsStepDetails],
            PropertyInfo(discriminator="type"),
        ]
    ]
    type: Literal["message_creation", "tool_calls"]



import pytest

from openai_responses import (
    AssistantsMock,
    ChatCompletionMock,
    EmbeddingsMock,
    FilesMock,
    ThreadsMock,
    MessagesMock,
    RunsMock,
    RunStepsMock,
)


def pytest_configure(config: pytest.Config):
    config.addinivalue_line(
        "markers",
        "assistants_mock: OpenAI assistants API mocker",
    )

    config.addinivalue_line(
        "markers",
        "chat_completion_mock: OpenAI chat completion API mocker",
    )

    config.addinivalue_line(
        "markers",
        "embeddings_mock: OpenAI embeddings API mocker",
    )

    config.addinivalue_line(
        "markers",
        "files_mock: OpenAI files API mocker",
    )

    config.addinivalue_line(
        "markers",
        "threads_mock: OpenAI threads API mocker",
    )

    config.addinivalue_line(
        "markers",
        "messages_mock: OpenAI messages API mocker",
    )

    config.addinivalue_line(
        "markers",
        "runs_mock: OpenAI runs API mocker",
    )

    config.addinivalue_line(
        "markers",
        "run_steps_mock: OpenAI runs steps API mocker",
    )


@pytest.fixture()
def assistants_mock() -> AssistantsMock:
    return AssistantsMock()


@pytest.fixture()
def chat_completion_mock() -> ChatCompletionMock:
    return ChatCompletionMock()


@pytest.fixture()
def embeddings_mock() -> EmbeddingsMock:
    return EmbeddingsMock()


@pytest.fixture()
def files_mock() -> FilesMock:
    return FilesMock()


@pytest.fixture()
def threads_mock() -> ThreadsMock:
    return ThreadsMock()


@pytest.fixture()
def messages_mock() -> MessagesMock:
    return MessagesMock()


@pytest.fixture()
def runs_mock() -> RunsMock:
    return RunsMock()


@pytest.fixture()
def run_steps_mock() -> RunStepsMock:
    return RunStepsMock()



import datetime as dt
from typing import Any, Optional, Type, TypeVar

from openai import BaseModel

M = TypeVar("M", bound=BaseModel)


def utcnow_unix_timestamp_s() -> int:
    return int(dt.datetime.now().timestamp())


def model_dict(m: BaseModel) -> dict[str, Any]:
    if hasattr(m, "model_dump"):
        return getattr(m, "model_dump")()
    else:
        return getattr(m, "dict")()


def model_parse(m: Type[M], d: Optional[object]) -> Optional[M]:
    if not d:
        return None
    if hasattr(m, "model_validate"):
        return getattr(m, "model_validate")(d)
    else:
        return getattr(m, "parse_obj")(d)
